// generated by nssdb/generate/main.go
package nssdb

var columns = map[string]string{
	"CKA_AC_ISSUER":                  "a83",
	"CKA_ALWAYS_AUTHENTICATE":        "a202",
	"CKA_ALWAYS_SENSITIVE":           "a165",
	"CKA_APPLICATION":                "a10",
	"CKA_ATTR_TYPES":                 "a85",
	"CKA_AUTH_PIN_FLAGS":             "a201",
	"CKA_BASE":                       "a132",
	"CKA_BITS_PER_PIXEL":             "a406",
	"CKA_CERTIFICATE_CATEGORY":       "a87",
	"CKA_CERTIFICATE_TYPE":           "a80",
	"CKA_CERT_MD5_HASH":              "ace5363b5",
	"CKA_CERT_SHA1_HASH":             "ace5363b4",
	"CKA_CHAR_COLUMNS":               "a404",
	"CKA_CHAR_ROWS":                  "a403",
	"CKA_CHAR_SETS":                  "a480",
	"CKA_CHECK_VALUE":                "a90",
	"CKA_CLASS":                      "a0",
	"CKA_COEFFICIENT":                "a128",
	"CKA_COLOR":                      "a405",
	"CKA_DECRYPT":                    "a105",
	"CKA_DEFAULT_CMS_ATTRIBUTES":     "a502",
	"CKA_DERIVE":                     "a10c",
	"CKA_EC_PARAMS":                  "a180",
	"CKA_EC_POINT":                   "a181",
	"CKA_ENCODING_METHODS":           "a481",
	"CKA_ENCRYPT":                    "a104",
	"CKA_END_DATE":                   "a111",
	"CKA_EXPONENT_1":                 "a126",
	"CKA_EXPONENT_2":                 "a127",
	"CKA_EXTRACTABLE":                "a162",
	"CKA_HASH_OF_ISSUER_PUBLIC_KEY":  "a8b",
	"CKA_HASH_OF_SUBJECT_PUBLIC_KEY": "a8a",
	"CKA_HAS_RESET":                  "a302",
	"CKA_HW_FEATURE_TYPE":            "a300",
	"CKA_ID":                         "a102",
	"CKA_ISSUER":                     "a81",
	"CKA_JAVA_MIDP_SECURITY_DOMAIN":  "a88",
	"CKA_KEY_GEN_MECHANISM":          "a166",
	"CKA_KEY_TYPE":                   "a100",
	"CKA_LABEL":                      "a3",
	"CKA_LOCAL":                      "a163",
	"CKA_MECHANISM_TYPE":             "a500",
	"CKA_MIME_TYPES":                 "a482",
	"CKA_MODIFIABLE":                 "a170",
	"CKA_MODULUS":                    "a120",
	"CKA_MODULUS_BITS":               "a121",
	"CKA_NEVER_EXTRACTABLE":          "a164",
	"CKA_NSS_DB":                     "ad5a0db00",
	"CKA_NSS_EMAIL":                  "ace534352",
	"CKA_NSS_EMAIL_DISTRUST_AFTER":   "ace534374",
	"CKA_NSS_EXPIRES":                "ace534357",
	"CKA_NSS_KRL":                    "ace534358",
	"CKA_NSS_MODULE_SPEC":            "ace534368",
	"CKA_NSS_OVERRIDE_EXTENSIONS":    "ace534369",
	"CKA_NSS_PASSWORD_CHECK":         "ace534356",
	"CKA_NSS_PKCS8_SALT":             "ace534355",
	"CKA_NSS_PQG_COUNTER":            "ace534364",
	"CKA_NSS_PQG_H":                  "ace534366",
	"CKA_NSS_PQG_SEED":               "ace534365",
	"CKA_NSS_PQG_SEED_BITS":          "ace534367",
	"CKA_NSS_SERVER_DISTRUST_AFTER":  "ace534373",
	"CKA_NSS_SMIME_INFO":             "ace534353",
	"CKA_NSS_SMIME_TIMESTAMP":        "ace534354",
	"CKA_NSS_TRUST":                  "a80000001",
	"CKA_NSS_URL":                    "ace534351",
	"CKA_OBJECT_ID":                  "a12",
	"CKA_OWNER":                      "a84",
	"CKA_PIXEL_X":                    "a400",
	"CKA_PIXEL_Y":                    "a401",
	"CKA_PRIME":                      "a130",
	"CKA_PRIME_1":                    "a124",
	"CKA_PRIME_2":                    "a125",
	"CKA_PRIME_BITS":                 "a133",
	"CKA_PRIVATE":                    "a2",
	"CKA_PRIVATE_EXPONENT":           "a123",
	"CKA_PUBLIC_EXPONENT":            "a122",
	"CKA_PUBLIC_KEY_INFO":            "a129",
	"CKA_REQUIRED_CMS_ATTRIBUTES":    "a501",
	"CKA_RESET_ON_INIT":              "a301",
	"CKA_RESOLUTION":                 "a402",
	"CKA_SECONDARY_AUTH":             "a200",
	"CKA_SENSITIVE":                  "a103",
	"CKA_SERIAL_NUMBER":              "a82",
	"CKA_SIGN":                       "a108",
	"CKA_SIGN_RECOVER":               "a109",
	"CKA_START_DATE":                 "a110",
	"CKA_SUBJECT":                    "a101",
	"CKA_SUBPRIME":                   "a131",
	"CKA_SUB_PRIME_BITS":             "a134",
	"CKA_SUPPORTED_CMS_ATTRIBUTES":   "a503",
	"CKA_TOKEN":                      "a1",
	"CKA_TRUSTED":                    "a86",
	"CKA_TRUST_CLIENT_AUTH":          "ace536359",
	"CKA_TRUST_CODE_SIGNING":         "ace53635a",
	"CKA_TRUST_CRL_SIGN":             "ace536357",
	"CKA_TRUST_DATA_ENCIPHERMENT":    "ace536354",
	"CKA_TRUST_DIGITAL_SIGNATURE":    "ace536351",
	"CKA_TRUST_EMAIL_PROTECTION":     "ace53635b",
	"CKA_TRUST_IPSEC_END_SYSTEM":     "ace53635c",
	"CKA_TRUST_IPSEC_TUNNEL":         "ace53635d",
	"CKA_TRUST_IPSEC_USER":           "ace53635e",
	"CKA_TRUST_KEY_AGREEMENT":        "ace536355",
	"CKA_TRUST_KEY_CERT_SIGN":        "ace536356",
	"CKA_TRUST_KEY_ENCIPHERMENT":     "ace536353",
	"CKA_TRUST_NON_REPUDIATION":      "ace536352",
	"CKA_TRUST_SERVER_AUTH":          "ace536358",
	"CKA_TRUST_STEP_UP_APPROVED":     "ace536360",
	"CKA_TRUST_TIME_STAMPING":        "ace53635f",
	"CKA_UNWRAP":                     "a107",
	"CKA_UNWRAP_TEMPLATE":            "a40000212",
	"CKA_URL":                        "a89",
	"CKA_VALUE":                      "a11",
	"CKA_VALUE_BITS":                 "a160",
	"CKA_VALUE_LEN":                  "a161",
	"CKA_VERIFY":                     "a10a",
	"CKA_VERIFY_RECOVER":             "a10b",
	"CKA_WRAP":                       "a106",
	"CKA_WRAP_TEMPLATE":              "a40000211",
	"CKA_WRAP_WITH_TRUSTED":          "a210",
}
